#include "extension.h"
#include <usercmd.h>

SH_DECL_MANUALHOOK5_void(ProcessUsercmds, 0, 0, 0, CUserCmd *, int, int, int, bool);

Sample g_Sample;		/**< Global singleton for extension's main interface */
SMEXT_LINK(&g_Sample);

IGameConfig *g_pGameConf = NULL;
int g_HookID = 0;

bool Sample::SDK_OnLoad(char *error, size_t maxlength, bool late)
{
	if (!gameconfs->LoadGameConfigFile("cmdfix.games", &g_pGameConf, error, maxlength))
		return false;

	int offset = 0;
	if (!g_pGameConf->GetOffset("CBasePlayer::ProcessUsercmds", &offset))
	{
		smutils->Format(error, maxlength, "Failed to find offset for CBasePlayer::ProcessUsercmds");
		return false;
	}
	SH_MANUALHOOK_RECONFIGURE(ProcessUsercmds, offset, 0, 0);

	playerhelpers->AddClientListener(&g_Sample);

	return true;
}

void Sample::SDK_OnUnload()
{
	gameconfs->CloseGameConfigFile(g_pGameConf);
	playerhelpers->RemoveClientListener(&g_Sample);

	if (g_HookID)
	{
		SH_REMOVE_HOOK_ID(g_HookID);
		g_HookID = 0;
	}
}

void Sample::OnClientPutInServer(int client)
{
	// We only want to hook once.
	if (g_HookID)
		return;

	IGamePlayer *pPlayer = playerhelpers->GetGamePlayer(client);
	CBaseEntity *pEntity = gamehelpers->ReferenceToEntity(client);

	if (!pPlayer || pPlayer->IsFakeClient() || !pEntity)
		return;

	g_HookID = SH_ADD_MANUALVPHOOK(ProcessUsercmds, pEntity, SH_MEMBER(&g_Sample, &Sample::Hook_ProcessUsercmds), false);
}

inline bool IsEntityCoordinateReasonable(const vec_t c)
{
	float r = 16384.0f; // k_flMaxEntityPosCoord
	return c > -r && c < r;
}

inline bool IsEntityQAngleReasonable(const QAngle &q)
{
	float r = 360.0 * 1000.0f; // k_flMaxEntityEulerAngle
	return
		q.x > -r && q.x < r &&
		q.y > -r && q.y < r &&
		q.z > -r && q.z < r;
}

bool IsUserCmdDataValid(CUserCmd *pCmd)
{
	return 
		// Prevent clients from sending invalid view angles to try to get leaf server code to crash
		(pCmd->viewangles.IsValid() && IsEntityQAngleReasonable(pCmd->viewangles)) &&
		// Movement ranges
		(IsFinite(pCmd->forwardmove) && IsEntityCoordinateReasonable(pCmd->forwardmove)) &&
		(IsFinite(pCmd->sidemove) && IsEntityCoordinateReasonable(pCmd->sidemove)) &&
		(IsFinite(pCmd->upmove) && IsEntityCoordinateReasonable(pCmd->upmove));
}

inline void MakeInert(CUserCmd *cmd)
{
	cmd->viewangles = QAngle(0, 0, 0); // vec3_angle
	cmd->forwardmove = 0.f;
	cmd->sidemove = 0.f;
	cmd->upmove = 0.f;
	cmd->buttons = 0;
	cmd->impulse = 0;
}

void Sample::Hook_ProcessUsercmds(CUserCmd *cmds, int numcmds, int totalcmds, int dropped_packets, bool paused)
{
	CBaseEntity *pEntity = META_IFACEPTR(CBaseEntity);
	IGamePlayer *pPlayer = playerhelpers->GetGamePlayer(gamehelpers->EntityToBCompatRef(pEntity));

	if (!pPlayer || pPlayer->IsFakeClient())
		RETURN_META(MRES_IGNORED);

	for (int i = totalcmds - 1; i >= 0; i--)
	{
		CUserCmd *pCmd = &cmds[totalcmds - 1 - i];

		// Validate values
		if (!IsUserCmdDataValid(pCmd))
		{
			MakeInert(pCmd);
/*
			if (!pPlayer->IsInKickQueue())
			{
				smutils->LogMessage(myself, "%s (ID: %s | IP: %s) was kicked for sending a corrupt user cmd.", pPlayer->GetName(), pPlayer->GetAuthString(false), pPlayer->GetIPAddress());
				gamehelpers->AddDelayedKick(pPlayer->GetIndex(), pPlayer->GetUserId(), "Corrupt user cmd");
			}
*/
		}
	}

	RETURN_META(MRES_IGNORED);
}